import grp
import pwd

from kisiac.common import HostAgnosticPath, log_action, run_cmd
from kisiac.config import Config


def setup_users(host: str) -> None:
    users = list(Config.get_instance().users)

    groups = {group for user in users for group in user.secondary_groups} | {
        user.primary_group for user in users
    }

    for group in groups:
        # create group if it does not exist
        if run_cmd(["getent", "group", group], check=False, host=host).returncode == 2:
            run_cmd(["groupadd", group], host=host, sudo=True)

    for user in Config.get_instance().users:
        # create user if it does not exist
        if not is_existing_user(user.username):
            group_arg = []
            if user.secondary_groups:
                group_arg = ["-G", ",".join(user.secondary_groups)]

            run_cmd(
                [
                    "useradd",
                    "-g",
                    user.primary_group,
                    *group_arg,
                    "--shell",
                    "/bin/bash",
                    "-m",
                    user.username,
                ],
                host=host,
                sudo=True,
            )
        else:
            log_action("Updating user", user.username)

        sshdir = HostAgnosticPath(f"~{user.username}/.ssh", host=host, sudo=True)
        sshdir.mkdir()
        sshdir.chown(user.username, user.primary_group)
        sshdir.chmod("u=rwx", "g-rwx", "o-rwx")
        auth_keys_file = sshdir / "authorized_keys"
        auth_keys_file.write_text(user.ssh_pub_key + "\n")
        user.fix_permissions([auth_keys_file.path], host=host)


def is_existing_user(username: str) -> bool:
    try:
        pwd.getpwnam(username)
        return True
    except KeyError:
        return False


def is_existing_group(groupname: str, host: str) -> bool:
    try:
        grp.getgrnam(groupname)
        return True
    except KeyError:
        return False



import subprocess as sp


def test_all():
    for i in range(2):
        print(f"Run {i}")
        sp.run(
            [
                "kisiac",
                "--non-interactive",
                "update-hosts",
                "--skip-system-upgrade",
                "localhost",
            ],
            check=True,
        )



from copy import copy
from dataclasses import dataclass
import json
from pathlib import Path
import re
from typing import Any, Self
from kisiac.common import HostAgnosticPath, UserError, confirm_action, run_cmd
from kisiac.config import Config, Filesystem, UserSet

from pyfstab import Fstab

blkid_attrs_re = re.compile(r'(?P<attr>[A-Z]+)="(?P<value>\S+)"')


def update_filesystems(host: str) -> None:
    filesystems = set(Config.get_instance().filesystems)
    device_infos = DeviceInfos(host)

    # First, create filesystems that do not exist yet or need to be changed.
    mkfs_cmds = []
    for filesystem in filesystems:
        device_info = device_infos.get_info(filesystem)
        if device_info.fstype != filesystem.fstype:
            mkfs_cmds.append(["mkfs", "-t", filesystem.fstype, str(device_info.device)])

    # Second, update /etc/fstab.
    fstab_path = HostAgnosticPath("/etc/fstab", host=host, sudo=True)
    old_fstab = Fstab().read_string(fstab_path.read_text())

    previous_entries = {
        Filesystem.from_fstab_entry(entry) for entry in old_fstab.entries
    }

    unchanged_entries = previous_entries & filesystems
    change_or_remove_msg = "\n".join(map(str, previous_entries - unchanged_entries))
    mkfs_cmds_msg = "\n".join(" ".join(cmd) for cmd in mkfs_cmds)

    if confirm_action(
        f"The following mkfs commands will be executed:\n{mkfs_cmds_msg}"
        f"\nThe following fstab entries will be changed or removed:\n{change_or_remove_msg}"
    ):
        for cmd in mkfs_cmds:
            run_cmd(cmd, sudo=True, host=host)

        new_fstab = Fstab()
        new_fstab.entries = [
            filesystem.to_fstab_entry() for filesystem in sorted(filesystems)
        ]
        fstab_path.write_text(new_fstab.write_string())


def update_permissions(host: str) -> None:
    def apply_user_set(user_set: UserSet | None, flag: str) -> list[str]:
        if user_set is None:
            return []
        if user_set == UserSet.owner:
            return [f"u+{flag}", f"g-{flag}", f"o-{flag}"]
        elif user_set == UserSet.group:
            return [f"u+{flag}", f"g+{flag}", f"o-{flag}"]
        elif user_set == UserSet.others:
            return [f"u+{flag}", f"g+{flag}", f"o+{flag}"]
        elif user_set == UserSet.nobody:
            return [f"u-{flag}", f"g-{flag}", f"o-{flag}"]
        else:
            assert False, "unreachable"

    permissions = Config.get_instance().permissions
    for path, permissions in permissions.items():
        path = HostAgnosticPath(path, host=host, sudo=True)
        chmod_args = []
        if permissions.setgid:
            chmod_args.append("g+s")
        if permissions.setuid:
            chmod_args.append("u+s")
        if permissions.sticky:
            chmod_args.append("+t")
        chmod_args.extend(apply_user_set(permissions.read, "r"))
        chmod_args.extend(apply_user_set(permissions.write, "w"))
        if chmod_args:
            path.chmod(*chmod_args)

        # execute permissions are handled differently for dir and files
        if path.is_dir():
            if permissions.read is not None:
                # With dirs, read should be considered equivalent to execute, and handled
                # non-recursively. In turn, we ignore the execute setting for dirs because
                # it becomes redundant.
                path.chmod(*apply_user_set(permissions.read, "x"), recursive=False)
        elif permissions.execute is not None:
            path.chmod(*apply_user_set(permissions.execute, "x"))
        path.chown(permissions.owner, permissions.group)


@dataclass
class DeviceInfo:
    device: Path
    fstype: str | None
    label: str | None
    uuid: str | None

    def is_targeted_by_filesystem(self, filesystem: Filesystem) -> bool:
        if filesystem.device is not None:
            return self.device == filesystem.device
        elif filesystem.label is not None:
            return self.label == filesystem.label
        elif filesystem.uuid is not None:
            return self.uuid == filesystem.uuid
        else:
            return False

    def with_device(self, device: Path) -> Self:
        info = copy(self)
        info.device = device
        return info


class DeviceInfos:
    def __init__(self, host: str) -> None:
        lsblk_output = json.loads(
            run_cmd(
                ["lsblk", "--json", "--path", "--fs"],
                sudo=True,
                host=host,
            ).stdout
        )
        self.infos: list[DeviceInfo] = []

        def parse_entry(entry: dict[str, Any]) -> None:
            device = Path(entry["name"])
            device_info = DeviceInfo(
                device=device,
                fstype=entry["fstype"],
                label=entry["label"],
                uuid=entry["uuid"],
            )
            self.infos.append(device_info)
            if device.is_relative_to(Path("/dev/mapper")):
                # also add /dev/vgname/lvname path for LVM logical volumes
                device = Path("/dev") / re.sub(
                    r"(?P<pre>[^-])-(?P<post>[^-])",
                    r"\g<pre>/\g<post>",
                    device.name,
                    count=1,
                ).replace("--", "-")
                self.infos.append(device_info.with_device(Path(device)))

            for child in entry.get("children", []):
                parse_entry(child)

        for entry in lsblk_output["blockdevices"]:
            parse_entry(entry)

    def get_info(self, filesystem: Filesystem) -> DeviceInfo:
        for info in self.infos:
            if info.is_targeted_by_filesystem(filesystem):
                return info
        raise UserError(
            f"No device found for filesystem with device={filesystem.device}, "
            f"label={filesystem.label}, uuid={filesystem.uuid}"
        )

    def get_info_for_device(self, device: Path) -> DeviceInfo:
        for info in self.infos:
            if info.device == device:
                return info
        raise UserError(f"No device info found for device {device}")



from pathlib import Path
import subprocess as sp
import sys
from typing import Any, Callable, Self, Sequence
import importlib
import re
import textwrap

import inquirer


cache = Path("~/.cache/kisiac").expanduser()


def handle_key_error(msg: str) -> Callable:
    def decoator(func: Callable) -> Callable:
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            try:
                return func(*args, **kwargs)
            except KeyError as e:
                raise UserError(f"{msg}: {e}") from e

        return wrapper

    return decoator


class Singleton:
    @classmethod
    def get_instance(cls, *args, **kwargs) -> Self:
        if not hasattr(cls, "_instance") or cls._instance is None:
            cls._instance = cls(*args, **kwargs)
        return cls._instance


def confirm_action(desc: str) -> bool:
    from kisiac.runtime_settings import GlobalSettings

    if GlobalSettings.get_instance().non_interactive:
        return True

    response = inquirer.prompt(
        [inquirer.Checkbox("action", message=desc, choices=["yes", "no"])]
    )
    assert response is not None
    return response["action"] == "yes"


def exists_cmd(cmd: str, host: str, sudo: bool) -> bool:
    try:
        run_cmd(["which", cmd], host=host, sudo=sudo, user_error=False)
        return True
    except sp.CalledProcessError:
        return False


def log_msg(*msgs: Any) -> None:
    print(" ".join(map(str, msgs)), file=sys.stderr)


def log_action(host: str, *msgs: Any) -> None:
    log_msg(f"[{host}]", *msgs)


def cmd_to_str(*cmds: list[str]) -> str:
    return "\n".join(" ".join(map(str, cmd)) for cmd in cmds)


def run_cmd(
    cmd: list[str],
    input: str | None = None,
    host: str = "localhost",
    env: dict[str, Any] | None = None,
    sudo: bool = False,
    user_error: bool = True,
    check: bool = True,
) -> sp.CompletedProcess[str]:
    """Run a system command using subprocess.run and check for errors."""
    # TODO check quotation!
    cmd = list(map(str, cmd))
    if sudo:
        cmd = ["sudo", "bash", "-c", f"{' '.join(cmd)}"]
    if host != "localhost":
        if sudo:
            cmd = ["ssh", host, f"sudo bash -c '{' '.join(cmd)}'"]
        else:
            cmd = ["ssh", host, f"{' '.join(cmd)}"]
    log_action(host, "Running command", cmd_to_str(cmd))
    try:
        return sp.run(
            cmd,
            check=check,
            text=True,
            stdout=sp.PIPE,
            stderr=sp.PIPE,
            input=input,
            env=env,
        )
    except sp.CalledProcessError as e:
        if user_error:
            raise UserError(
                f"Error occurred while running command '{' '.join(cmd)}': {e.stderr}"
            ) from e
        else:
            raise


class UserError(Exception):
    """Base class for user-related errors."""

    pass


def check_type(item: str, value: Any, expected_type: Any) -> None:
    if not isinstance(value, expected_type):
        raise UserError(
            f"Expecting {expected_type} for {item}, found "
            f"value {value} of type {type(value)}."
        )


module_import_re: re.Pattern = re.compile(
    r"from kisiac.(?P<module>[a-z0-9_]+) import [a-zA-Z0-0_,+]"
)


def func_to_sh(func: Callable) -> str:
    func_name = func.__name__
    module_code = get_module_code(func.__module__)
    return textwrap.dedent(f"""
    {module_code}
    {func_name}()
    """)


def get_module_code(module_name: str) -> str:
    module_path = importlib.import_module(module_name).__file__
    assert module_path is not None

    with open(module_path, mode="r") as f:
        module_code = module_import_re.sub(
            lambda match: get_module_code(f"kisiac.{match.group('module')}"), f.read()
        )

    return module_code


class HostAgnosticPath:
    def __init__(
        self, path: str | Path, host: str = "localhost", sudo: bool = False
    ) -> None:
        self.path = Path(path)
        self.host = host
        self.sudo = sudo
        if self.is_local_and_user():
            # if non local or sudo, shell commands will be used, which expand
            # the ~ operator automatically
            self.path = self.path.expanduser()

    def read_text(self) -> str:
        if self.is_local_and_user():
            return self.path.read_text()
        else:
            return self._run_cmd(["cat", str(self.path)]).stdout

    def write_text(self, content: str) -> None:
        if self.is_local_and_user():
            self.path.write_text(content)
        else:
            self._run_cmd(
                ["tee", str(self.path)],
                input=content,
            )

    def mkdir(self) -> None:
        if self.is_local_and_user():
            self.path.mkdir(parents=True, exist_ok=True)
        else:
            self._run_cmd(["mkdir", "-p", str(self.path)])

    def chmod(self, *mode: str, recursive: bool = True) -> None:
        self._chperm("chmod", ",".join(mode), recursive=recursive)

    def chown(self, user: str | None, group: str | None = None) -> None:
        if user is not None:
            owner = f"{user}:{group}" if group else user
            self._chperm("chown", owner)
        elif group is not None:
            self._chperm("chgrp", group)
        else:
            raise ValueError("Either user or group must be provided.")

    def _chperm(self, cmd: str, arg: str, recursive: bool = True) -> None:
        args = [arg]
        if recursive and self.is_dir():
            args = ["-R", *args]
        self._run_cmd([cmd, *args, str(self.path)])

    def is_local_and_user(self) -> bool:
        return self.host == "localhost" and not self.sudo

    def _run_cmd(
        self, cmd: list[str], input: str | None = None, user_error: bool = True
    ) -> sp.CompletedProcess[str]:
        return run_cmd(
            cmd,
            input=input,
            host=self.host,
            sudo=self.sudo,
            user_error=user_error,
        )

    def exists(self) -> bool:
        if self.is_local_and_user():
            return self.path.exists()
        else:
            try:
                self._run_cmd(
                    ["test", "-e", str(self.path)],
                    user_error=False,
                )
                return True
            except sp.CalledProcessError:
                return False

    def is_dir(self) -> bool:
        if self.is_local_and_user():
            return self.path.is_dir()
        else:
            try:
                self._run_cmd(
                    ["test", "-d", str(self.path)],
                    user_error=False,
                )
                return True
            except sp.CalledProcessError:
                return False

    def with_suffix(self, suffix: str) -> Self:
        return type(self)(self.path.with_suffix(suffix), host=self.host, sudo=self.sudo)

    @property
    def parents(self) -> Sequence[Self]:
        return [
            type(self)(parent, host=self.host, sudo=self.sudo)
            for parent in self.path.parents
        ]

    def __truediv__(self, other: Any) -> Self:
        return type(self)(self.path / other, host=self.host, sudo=self.sudo)

    def __rtruediv__(self, other: Any) -> Self:
        return type(self)(other / self.path, host=self.host, sudo=self.sudo)

    def __str__(self) -> str:
        if self.host == "localhost":
            return str(self.path)
        else:
            return f"{self.host}:{self.path}"



from dataclasses import dataclass, field, fields
from argparse import ArgumentParser, Namespace
from typing import Self, get_args, get_origin

from kisiac.common import Singleton


@dataclass
class SettingsBase(Singleton):
    @classmethod
    def register_cli_args(cls, parser: ArgumentParser) -> None:
        for cls_field in fields(cls):
            positional = cls_field.metadata.get("positional", False)

            arg_name = cls_field.name.replace("_", "-")

            parse_method = getattr(cls, f"parse_{cls_field.name}", None)
            arg_type = parse_method or cls_field.type

            default = None
            if callable(cls_field.default_factory):
                default = cls_field.default_factory()
            elif cls_field.default is not None:
                default = cls_field.default

            kwargs = dict(
                help=cls_field.metadata["help"],
            )
            if cls_field.metadata.get("required", False) and not positional:
                kwargs["required"] = True

            if arg_type is bool:
                kwargs["action"] = "store_true" if not default else "store_false"
            else:
                kwargs["default"] = default
                if cls_field.type is list or get_origin(cls_field.type) is list:
                    kwargs["type"] = get_args(cls_field.type)[0]
                    kwargs["nargs"] = "+"
                else:
                    kwargs["type"] = arg_type

            metavar = cls_field.metadata.get("metavar", None)
            if metavar is not None:
                kwargs["metavar"] = metavar

            parser.add_argument(
                f"--{arg_name}" if not positional else arg_name,
                **kwargs,
            )

    @classmethod
    def from_cli_args(cls, args: Namespace) -> Self:
        kwargs = {
            cls_field.name: getattr(args, cls_field.name) for cls_field in fields(cls)
        }
        return cls.get_instance(**kwargs)


@dataclass
class GlobalSettings(SettingsBase):
    non_interactive: bool = field(
        default=False, metadata={"help": "Run in non-interactive mode"}
    )


@dataclass
class UpdateHostSettings(SettingsBase):
    skip_system_upgrade: bool = field(
        default=False,
        metadata={"help": "Skip system package upgrades"},
    )
    hosts: list[str] = field(
        default_factory=lambda: ["localhost"],
        metadata={
            "required": True,
            "positional": True,
            "help": "Hosts to update (default: localhost)",
            "metavar": "HOST",
        },
    )



from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser

from kisiac.common import UserError, log_msg
from kisiac.runtime_settings import (
    GlobalSettings,
    UpdateHostSettings,
)
from kisiac.update import setup_config, update_host


def get_argument_parser() -> ArgumentParser:
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    GlobalSettings.register_cli_args(parser)
    subparsers = parser.add_subparsers(dest="subcommand", help="subcommand help")
    update_hosts = subparsers.add_parser(
        "update-hosts",
        help="Update given hosts",
        description="Update given hosts",
        formatter_class=ArgumentDefaultsHelpFormatter,
    )
    UpdateHostSettings.register_cli_args(update_hosts)
    subparsers.add_parser(
        "setup-config",
        help="Setup the kisiac configuration",
        description="Setup the kisiac configuration",
    )

    return parser


def main() -> None:
    try:
        parser = get_argument_parser()
        args = parser.parse_args()
        GlobalSettings.from_cli_args(args)
        match args.subcommand:
            case "update-hosts":
                UpdateHostSettings.from_cli_args(args)
                for host in UpdateHostSettings.get_instance().hosts:
                    update_host(host)
            case "setup-config":
                setup_config()
            case _:
                parser.print_help()
    except UserError as e:
        log_msg(e)
        exit(1)
